﻿Imports C1.Win.C1Command

Namespace CRUD
    Public Class ControlMenuCommand
        Implements IDisposable

        ''' <summary>
        ''' The name of the command, with accessibility chars.
        ''' </summary>
        Public ReadOnly Name As String

        ''' <summary>
        ''' The name of the command, sans accessibility chars.
        ''' </summary>
        Public ReadOnly CleanName As String

        ''' <summary>
        ''' Enables or Disables the menu command.
        ''' </summary>
        Public Property Enabled As Boolean
            Get
                Return mCommand.Enabled
            End Get
            Set(value As Boolean)
                mCommand.Enabled = value
            End Set
        End Property

        ''' <summary>
        ''' Gets the image associated with this command.
        ''' </summary>
        Public ReadOnly Property Image As Drawing.Image
            Get
                Return mCommand.Image
            End Get
        End Property

        ''' <summary>
        ''' Gets the shortcut associated with this command.
        ''' </summary>
        Public ReadOnly Property Shortcut As Shortcut
            Get
                Return mCommand.Shortcut
            End Get
        End Property

        ''' <summary>
        ''' If this is a submenu command, returns the parent command; otherwise null.
        ''' </summary>
        Public ReadOnly Property ParentCommand As ControlMenuCommand
            Get
                Return mParentCommand
            End Get
        End Property

        ''' <summary>
        ''' Whether the command responds to mouse clicks. Returns false if has children; true otherwise.
        ''' </summary>
        Public ReadOnly Property IsClickable As Boolean
            Get
                Return Not TypeOf mCommand Is C1CommandMenu
            End Get
        End Property

        ''' <summary>
        ''' A payload of some extra data that might be used by a strategy, form, etc.
        ''' </summary>
        Public Property UserData As Object

        Public ReadOnly IsAvailableInSupplementaryCommandsRibbonMenu As Boolean

        'Used when swapping the command.
        Private mCommandHolder As C1CommandHolder

        'The command's owner menu, if we remove it from here we remove it from every other menu.
        Private mPrimaryMenu As C1ContextMenu

        'The commands parent, if submenu command.
        Private mParentCommand As ControlMenuCommand

        'Our precious.
        Private mCommand As C1Command

        'Every command link that references our command.
        Private mCommandLinks As List(Of WeakReference(Of C1CommandLink))

        'Every child this command has.
        Private ReadOnly mChildCommands As List(Of ControlMenuCommand)

        ''' <summary>
        ''' Fires on command mouse/alt click.
        ''' </summary>
        Friend Event OnClick(Sender As ControlMenuCommand)
        Public Event OnDisposed(Sender As ControlMenuCommand)
        ''' <summary>
        ''' Creates a new wrapper for a C1Command in order to keep all menus that share this command in sync. 
        ''' The underlying C1Command should not circumstances be exposed, or we'll break the aforementioned encapsulation. 
        ''' </summary>
        ''' <param name="PrimaryMenu">Considered the owner menu of this command.</param>
        Friend Sub New(CommandHolder As C1CommandHolder,
                       PrimaryMenu As C1ContextMenu,
                       Name As String,
                       Image As Drawing.Image,
                       Shortcut As Shortcut,
                       Optional AddToSupplementaryCommandsRibbonMenu As Boolean = False,
                       Optional ControlName As String = Nothing)

            mCommandHolder = CommandHolder
            mPrimaryMenu = PrimaryMenu
            mParentCommand = Nothing
            mCommand = mCommandHolder.CreateCommand
            IsAvailableInSupplementaryCommandsRibbonMenu = AddToSupplementaryCommandsRibbonMenu

            AddHandler mCommand.Click, AddressOf EventCommandClick

            With mCommand
                .Text = Name
                .Image = Image

                'We'll give C1 a bogus shortcut, so it doesn't trigger on our real shortcuts.
                If Shortcut <> Windows.Forms.Shortcut.None Then
                    .ShowShortcut = True
                    .Shortcut = Shortcut.Alt0
                    .ShortcutText = Text.RegularExpressions.Regex.Replace(Shortcut.ToString, "([a-z])([A-Z0-9])", "$1+$2")

                Else
                    .ShowShortcut = False
                    .Shortcut = Windows.Forms.Shortcut.None
                End If
            End With

            Me.Name = Name
            Me.CleanName = GetCleanName(Name)
            Me.ControlName = ControlName

            mCommandLinks = New List(Of WeakReference(Of C1CommandLink))
            mChildCommands = New List(Of ControlMenuCommand)
        End Sub

        Public Property ControlName As String

        ''' <summary>
        ''' Get the list of child commands if any
        ''' </summary>
        Friend ReadOnly Property HasChildren() As Boolean
            Get
                Return mChildCommands.Count > 0
            End Get
        End Property

        ''' <summary>
        ''' Get the list of child commands if any
        ''' </summary>
        Friend ReadOnly Property ChildCommands() As List(Of ControlMenuCommand)
            Get
                Return mChildCommands
            End Get
        End Property

        ''' <summary>
        ''' Property to get the command object
        ''' </summary>
        Friend ReadOnly Property Command() As C1Command
            Get
                Return mCommand
            End Get
        End Property

        ''' <summary>
        ''' Whether this is the owner menu of this command.
        ''' </summary>
        Friend Function IsPrimaryMenu(Menu As C1ContextMenu) As Boolean
            Return Menu Is mPrimaryMenu
        End Function

        Friend ReadOnly Iterator Property GetChildCommandNames() As IEnumerable(Of String)
            Get
                For Each lChildCommand In mChildCommands.ToArray
                    Yield lChildCommand.CleanName
                Next
            End Get
        End Property

        ''' <summary>
        ''' Associates this command with a context menu or a command that belongs to that context menu.
        ''' </summary>
        Friend Sub AddToMenu(CommandLinks As C1CommandLinks, IsTopDelimited As Boolean, Optional InsertAbove As String = Nothing)
            Dim lCommandLink As New C1CommandLink(mCommand) With {.Delimiter = IsTopDelimited}
            Dim lInsertIndex As Integer

            'We want to insert that command at a specific position
            If InsertAbove IsNot Nothing Then
                lInsertIndex = GetCommandLinkIndex(CommandLinks, InsertAbove)
            Else
                lInsertIndex = CommandLinks.Count
            End If

            mCommandLinks.Add(New WeakReference(Of C1CommandLink)(lCommandLink))
            CommandLinks.Insert(lInsertIndex, lCommandLink)
        End Sub

        ''' <summary>
        ''' Disassociates this command from a context menu. If this is the owner menu, disassociate from every other menu as well.
        ''' </summary>
        Friend Sub RemoveFromMenu(Menu As C1ContextMenu, CommandLinks As C1CommandLinks)
            Dim lCommandLink As C1CommandLink = Nothing

            For Each lRow In mCommandLinks.ToArray
                'If the weak reference is dead remove it and move on.
                If Not lRow.TryGetTarget(lCommandLink) Then
                    mCommandLinks.Remove(lRow)
                    Continue For
                End If

                'Either belongs to the primary menu or the specific menu, either way we need to remove.
                If Menu Is mPrimaryMenu OrElse CommandLinks.Contains(lCommandLink) Then
                    DestroyCommandLink(lCommandLink)
                    mCommandLinks.Remove(lRow)
                End If
            Next
        End Sub

        ''' <summary>
        ''' Adds a sub-menu command to the current menu.
        ''' </summary>
        Friend Sub AddChild(Command As ControlMenuCommand, IsTopDelimited As Boolean, Optional InsertAbove As String = Nothing)
            'We'll need to turn our vanilla command to a command menu, if it's not one already.
            If Not TypeOf mCommand Is C1CommandMenu Then
                SwapUnderlyingCommand(New C1CommandMenu)
            End If

            Command.AddToMenu(DirectCast(mCommand, C1CommandMenu).CommandLinks, IsTopDelimited, InsertAbove)
            Command.SetParentCommand(Me)
            mChildCommands.Add(Command)
        End Sub

        ''' <summary>
        ''' Removes a sub-menu command to the current menu.
        ''' </summary>
        Friend Sub RemoveChild(Command As ControlMenuCommand)
            Command.RemoveFromMenu(mPrimaryMenu, DirectCast(mCommand, C1CommandMenu).CommandLinks)
            Command.SetParentCommand(Nothing)
            mChildCommands.Remove(Command)

            'We have no more children, time to turn the command menu to a vanilla command.
            If mChildCommands.Count = 0 Then
                SwapUnderlyingCommand(mCommandHolder.CreateCommand)
            End If
        End Sub

        ''' <summary>
        ''' Used by <see cref="AddChild" /> and <see cref="RemoveChild" /> on the child command.
        ''' </summary>
        Friend Sub SetParentCommand(Command As ControlMenuCommand)
            mParentCommand = Command
        End Sub

#Region "Helpers"
        'Strips out the accessibility character from the command text.
        Private Function GetCleanName(CommandText As String) As String
            Return CommandText.Replace("&", "")
        End Function

        'Gets a command's ordinal in a set of command links.
        Private Function GetCommandLinkIndex(CommandLinks As C1CommandLinks, Name As String) As Integer
            For i = 0 To CommandLinks.Count - 1
                If GetCleanName(CommandLinks(i).Command.Text) = GetCleanName(Name) Then
                    Return i
                End If
            Next

            Throw New KeyNotFoundException(String.Format("Cannot find the index of command ""{0}"".", Name))
        End Function

        'Removes the command link from its owner and disposes it.
        Private Sub DestroyCommandLink(CommandLink As C1CommandLink)
            If TypeOf CommandLink.Owner Is C1ContextMenu Then
                DirectCast(CommandLink.Owner, C1ContextMenu).CommandLinks.Remove(CommandLink)

            ElseIf TypeOf CommandLink.Owner Is C1CommandMenu Then
                DirectCast(CommandLink.Owner, C1CommandMenu).CommandLinks.Remove(CommandLink)

            Else
                Throw New NotImplementedException(String.Format("Can't remove command from owner of type ""{0}"".", CommandLink.Owner.GetType.Name))
            End If

            CommandLink.Dispose()
            CommandLink = Nothing
        End Sub

        'Swaps one command for another. Duplicates all properties from the old, and replaces the command links.
        Private Sub SwapUnderlyingCommand(NewCommand As C1Command)
            Dim lOldCommand = mCommand
            Dim lCommandLink As C1CommandLink = Nothing

            'Replicate all command properties from the old command
            With NewCommand
                .Text = lOldCommand.Text
                .Image = lOldCommand.Image
                .Shortcut = lOldCommand.Shortcut
                .ShortcutText = lOldCommand.ShortcutText
                .ShowShortcut = Not TypeOf NewCommand Is C1CommandMenu
            End With

            'And go through every command link, replacing the old command with the new one.
            For Each lRow In mCommandLinks
                If lRow.TryGetTarget(lCommandLink) Then
                    lCommandLink.Command = NewCommand
                End If
            Next

            Dim lFoundIndex As Integer = GetCommandReferenceIndex(mCommand)
            If GetCommandReferenceIndex(mCommand) >= 0 Then
                mCommandHolder.Commands.RemoveAt(lFoundIndex)
            End If

            'Swap the old command instance
            mCommand = NewCommand

            'After all the links have the new command, we can dispose of the old one.
            lOldCommand.Dispose()
        End Sub

        Private Function GetCommandReferenceIndex(commandObj As C1Command) As Integer
            For lIndex As Integer = 0 To mCommandHolder.Commands.Count - 1
                If commandObj.GetHashCode = mCommandHolder.Commands(lIndex).GetHashCode Then
                    Return lIndex
                End If
            Next

            Return -1
        End Function

        'Tell the outside world something happened
        Friend Sub RaiseOnClick()
            RaiseEvent OnClick(Me)
        End Sub

        Private Sub EventCommandClick(sender As Object, e As ClickEventArgs)
            RaiseOnClick()
        End Sub
#End Region

#Region "IDisposable"
        Private mIsDisposed As Boolean

        Protected Overridable Sub Dispose(Disposing As Boolean)
            If Not mIsDisposed Then
                If Disposing Then
                    Dim lCommandLink As C1CommandLink = Nothing

                    'Anyone menu that still uses this command is SOL.
                    For Each lRow In mCommandLinks
                        If lRow.TryGetTarget(lCommandLink) Then
                            DestroyCommandLink(lCommandLink)
                        End If
                    Next

                    'And all child commands are going down as well.
                    For Each lChildCommand In mChildCommands
                        lChildCommand.Dispose()
                    Next

                    For lIndex As Integer = 0 To mCommandHolder.Commands.Count - 1
                        If mCommandHolder.Commands(lIndex).GetHashCode() = mCommand.GetHashCode() Then
                            mCommandHolder.Commands.RemoveAt(lIndex)
                            Exit For
                        End If
                    Next

                    If mCommand IsNot Nothing Then
                        mCommand.Image = Nothing
                        RemoveHandler mCommand.Click, AddressOf EventCommandClick
                        mCommand.Dispose()
                        mCommand = Nothing
                    End If

                    If mCommandHolder IsNot Nothing Then
                        mCommandHolder = Nothing
                    End If

                    If mPrimaryMenu IsNot Nothing Then
                        If mPrimaryMenu.C1ContextMenu IsNot Nothing Then
                            mPrimaryMenu.C1ContextMenu.Dispose()
                            mPrimaryMenu.C1ContextMenu = Nothing
                        End If
                        mPrimaryMenu = Nothing
                    End If

                    mCommandLinks.Clear()
                    mChildCommands.Clear()
                End If
            End If

            RaiseEvent OnDisposed(Me)
            mIsDisposed = True
        End Sub

        Public Sub Dispose() Implements IDisposable.Dispose
            Dispose(True)
            GC.SuppressFinalize(Me)
        End Sub
#End Region
    End Class
End Namespace