﻿using System;
using System.Drawing;
using System.ComponentModel;
using System.ComponentModel.Design;
using System.Windows.Forms;
using System.Runtime.InteropServices;

namespace Eka.Inheritance.MDIControls
{
    /// <summary>
    /// Listens for messages sent to a <see cref="System.Windows.Forms.MdiClient"/>
    /// class and controls its properties.
    /// </summary>	
    public class MdiClientController : NativeWindow, IComponent, IDisposable
    {
        #region Private Members
        /// <summary>
        /// MDI form where the Style has to be applied
        /// </summary>
        private Form mParentMDIForm;

        /// <summary>
        /// MdiClient, responsible to change the style of the MDI Form
        /// </summary>
        private MdiClient mMdiClient;

        /// <summary>
        /// MDI Form's Border style
        /// </summary>
        private BorderStyle mBorderStyle;

        /// <summary>
        /// MDI Form's BackColor
        /// </summary>
        private Color mBackColor;

        /// <summary>
        /// MDI Form's AutoScrolll Properties
        /// </summary>
        private bool mAutoScroll;

        /// <summary>
        /// MDI Form's Image to be set if any
        /// </summary>
        private Image mImage;

        /// <summary>
        /// MDI Form's Image Alignment
        /// </summary>
        private ContentAlignment mImageAlignment;

        /// <summary>
        /// MDI Form's image to be stretch or not
        /// </summary>
        private bool stretchImage;

        /// <summary>
        /// MDI Form's ISite
        /// </summary>
        private ISite mSite;
        #endregion

        #region Events
        /// <summary>
        /// Occurs when the control is redrawn.
        /// </summary>
        [Category("Appearance"), Description("Occurs when a control needs repainting.")]
        public event PaintEventHandler Paint;

        /// <summary>
        /// Occurs when the control is disposed.
        /// </summary>
        [Browsable(false)]
        public event EventHandler Disposed;

        /// <summary>
        /// Occurs when the <see cref="System.Windows.Forms.NativeWindow"/> handle
        /// is assigned.
        /// </summary>
        [Browsable(false)]
        public event EventHandler HandleAssigned;
        #endregion

        #region Constructors
        /// <summary>
        /// Initializes a new instance of the <see cref="MdiClientController"/> class.
        /// </summary>
        public MdiClientController()
            : this(null)
        {
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="MdiClientController"/> class
        /// for the given MDI form.
        /// </summary>
        /// <param name="parentForm">The MDI form.</param>
        public MdiClientController(Form parentForm)
        {
            // Initialize the default variables.
            this.mSite = null;
            this.mParentMDIForm = null;
            this.mMdiClient = null;
            this.mBackColor = SystemColors.AppWorkspace;
            this.mBorderStyle = BorderStyle.Fixed3D;
            this.mAutoScroll = true;
            this.mImage = null;
            this.mImageAlignment = ContentAlignment.MiddleCenter;
            this.stretchImage = false;

            // Set the ParentForm property.
            this.ParentForm = parentForm;
        }
        #endregion

        #region Public Properties
        /// <summary>
        /// Gets or sets the <see cref="System.ComponentModel.ISite"/> of
        /// the <see cref="System.ComponentModel.Component"/>.
        /// </summary>
        [Browsable(false)]
        [DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
        public ISite Site
        {
            get
            {
                return mSite;
            }

            set
            {
                mSite = value;

                if (mSite == null)
                    return;

                // If the component is dropped onto a form during design-time,
                // set the ParentForm property.
                IDesignerHost lHost = (value.GetService(typeof(IDesignerHost)) as IDesignerHost);

                if (lHost != null)
                {
                    Form lParent = lHost.RootComponent as Form;
                    if (lParent != null)
                        ParentForm = lParent;
                }
            }
        }

        /// <summary>
        /// Gets or sets the form that the <see cref="System.Windows.Forms.MdiClient"/>
        /// control is assigned to.
        /// </summary>
        [Browsable(false)]
        public Form ParentForm
        {
            get
            {
                return mParentMDIForm;
            }

            set
            {
                mParentMDIForm = value;

                if (mParentMDIForm == null)
                    return;

                // If the ParentForm has previously been set, unwire events connected to the old parent.					
                mParentMDIForm.HandleCreated -= new EventHandler(ParentFormHandleCreated);

                // If the parent form has not been created yet,
                // wait to initialize the MDI client until it is.
                if (mParentMDIForm.IsHandleCreated)
                {
                    InitializeMdiClient();
                    RefreshProperties();
                }
                else
                {
                    mParentMDIForm.HandleCreated += new EventHandler(ParentFormHandleCreated);
                }
            }
        }

        /// <summary>
        /// Gets the <see cref="System.Windows.Forms.MdiClient"/> being controlled.
        /// </summary>
        [Browsable(false)]
        public MdiClient MdiClient
        {
            get
            {
                return mMdiClient;
            }
        }

        /// <summary>
        /// Gets or sets the background color for the control.
        /// </summary>
        [Category("Appearance"), DefaultValue(typeof(Color), "AppWorkspace")]
        [Description("The backcolor used to display text and graphics in the control.")]
        public Color BackColor
        {
            // Use the BackColor property of the MdiClient control. This is one of
            // the few properties in the MdiClient class that actually works.
            get
            {
                if (mMdiClient != null)
                    return mMdiClient.BackColor;

                return mBackColor;
            }

            set
            {
                mBackColor = value;

                if (mMdiClient != null)
                    mMdiClient.BackColor = value;
            }
        }

        /// <summary>
        /// Indicates the border style for the control.
        /// </summary>
        /// <exception cref="System.ComponentModel.InvalidEnumArgumentException">The value
        /// assigned to the property is not a member of
        /// the <see cref="System.Windows.Forms.BorderStyle"/> enumeration.</exception>
        [DefaultValue(BorderStyle.Fixed3D), Category("Appearance")]
        [Description("Indicates whether the MDI area should have a border.")]
        public BorderStyle BorderStyle
        {
            get
            {
                return mBorderStyle;
            }

            set
            {
                // Error-check the enum.
                if (!Enum.IsDefined(typeof(BorderStyle), value))
                    throw new InvalidEnumArgumentException("value", (int)value, typeof(BorderStyle));

                mBorderStyle = value;

                if (mMdiClient == null)
                    return;

                // This property can actually be visible in design-mode,
                // but to keep it consistent with the others, prevent this from being show at design-time.
                if (mSite != null && mSite.DesignMode)
                    return;

                // There is no BorderStyle property exposed by the MdiClient class,
                // but this can be controlled by Win32 functions. A Win32 ExStyle
                // of WS_EX_CLIENTEDGE is equivalent to a Fixed3D border and a
                // Style of WS_BORDER is equivalent to a FixedSingle border.
                // Get styles using Win32 calls
                int style = GetWindowLong(mMdiClient.Handle, GWL_STYLE);
                int exStyle = GetWindowLong(mMdiClient.Handle, GWL_EXSTYLE);

                // Add or remove style flags as necessary.
                switch (mBorderStyle)
                {
                    case BorderStyle.Fixed3D:
                        exStyle |= WS_EX_CLIENTEDGE;
                        style &= ~WS_BORDER;
                        break;

                    case BorderStyle.FixedSingle:
                        exStyle &= ~WS_EX_CLIENTEDGE;
                        style |= WS_BORDER;
                        break;

                    case BorderStyle.None:
                        style &= ~WS_BORDER;
                        exStyle &= ~WS_EX_CLIENTEDGE;
                        break;
                }

                // Set the styles using Win32 calls
                SetWindowLong(mMdiClient.Handle, GWL_STYLE, style);
                SetWindowLong(mMdiClient.Handle, GWL_EXSTYLE, exStyle);

                // Cause an update of the non-client area.
                UpdateStyles();
            }
        }

        /// <summary>
        /// Gets or sets a value indicating whether the container will allow the user to
        /// scroll to any controls placed outside of its visible boundaries.
        /// </summary>
        [DefaultValue(true), Category("Layout")]
        [Description("Determines whether scrollbars will automatically appear if controls are placed outside the MDI client area.")]
        public bool AutoScroll
        {
            get
            {
                return mAutoScroll;
            }

            set
            {
                // By default the MdiClient control scrolls. It can appear though that
                // there are no scrollbars by turning them off when the non-client area is calculated. 
                mAutoScroll = value;
                if (mMdiClient != null)
                    UpdateStyles();
            }
        }

        /// <summary>
        /// Gets or sets the image that the
        /// <see cref="MdiClientController"/> displays.
        /// </summary>
        [DefaultValue(null), Category("Appearance")]
        [Description("The image displayed in the MDI client area.")]
        public Image Image
        {
            get
            {
                return mImage;
            }

            set
            {
                mImage = value;
                if (mMdiClient != null)
                    mMdiClient.Invalidate();
            }
        }

        /// <summary>
        /// Gets or sets the alignment of the background image.
        /// </summary>
        /// <exception cref="System.ComponentModel.InvalidEnumArgumentException">The
        /// value assigned to the property is not a member of
        /// the <see cref="System.Drawing.ContentAlignment"/> enumeration.</exception>
        [Category("Appearance"), DefaultValue(ContentAlignment.MiddleCenter)]
        [Description("Determines the position of the image within the MDI client area.")]
        public ContentAlignment ImageAlign
        {
            get
            {
                return mImageAlignment;
            }

            set
            {
                // Error-check the enum.
                if (!Enum.IsDefined(typeof(ContentAlignment), value))
                    throw new InvalidEnumArgumentException("value", (int)value, typeof(ContentAlignment));

                mImageAlignment = value;
                if (mMdiClient != null)
                    mMdiClient.Invalidate();
            }
        }

        /// <summary>
        /// Gets or sets a value indicating whether the
        /// <see cref="Image"/> should be stretched to fill the entire client area.
        /// </summary>
        [Category("Appearance"), DefaultValue(false)]
        [Description("Determines whether the image should be scaled to fill the entire client area.")]
        public bool StretchImage
        {
            get
            {
                return stretchImage;
            }

            set
            {
                stretchImage = value;
                if (mMdiClient != null)
                    mMdiClient.Invalidate();
            }
        }

        /// <summary>
        /// Gets the handle for this window.
        /// </summary>
        [Browsable(false)]
        public new IntPtr Handle
        {
            // Hide this from the property grid during design-time.
            get
            {
                return base.Handle;
            }
        }

        #endregion

        #region Public Methods
        /// <summary>
        /// Releases all resources used by the
        /// <see cref="System.ComponentModel.Component"/>.
        /// </summary>
        public void Dispose()
        {
            Dispose(true);
            GC.SuppressFinalize(this);
        }

        /// <summary>
        /// Reestablishes a connection to the <see cref="System.Windows.Forms.MdiClient"/>
        /// control if the <see cref="MdiClientController.ParentForm"/>
        /// hasn't changed but its <see cref="System.Windows.Forms.Form.IsMdiContainer"/>
        /// property has.
        /// </summary>
        public void RenewMdiClient()
        {
            // Reinitialize the MdiClient and its properties.
            InitializeMdiClient();
            RefreshProperties();
        }
        #endregion

        #region Protected Methods
        /// <summary>Releases the unmanaged resources used by the
        /// <see cref="System.ComponentModel.Component"/> and optionally releases the
        /// managed resources.</summary>
        /// <param name="disposing"><c>true</c> to release both managed and unmanaged
        /// resources; <c>false</c> to release only unmanaged resources.</param>
        protected virtual void Dispose(bool disposing)
        {
            if (disposing)
            {
                lock (this)
                {
                    if (mSite != null && mSite.Container != null)
                        mSite.Container.Remove(this);

                    if (Disposed != null)
                        Disposed(this, EventArgs.Empty);
                }
            }
        }

        /// <summary>
        /// Invokes the default window procedure associated with this window.
        /// </summary>
        /// <param name="Msg">A <see cref="System.Windows.Forms.Message"/> that is associated with the current Windows message. </param>
        protected override void WndProc(ref Message Msg)
        {
            switch (Msg.Msg)
            {
                //Do all painting in WM_PAINT to reduce flicker.
                case WM_ERASEBKGND:
                    return;

                case WM_PAINT:
                    // Use Win32 to get a Graphics object.
                    PAINTSTRUCT lPaintStruct = new PAINTSTRUCT();
                    IntPtr lScreenHdc = BeginPaint(Msg.HWnd, ref lPaintStruct);

                    using (Graphics screenGraphics = Graphics.FromHdc(lScreenHdc))
                    {
                        // Get the area to be updated.
                        Rectangle clipRect = new Rectangle(
                            lPaintStruct.rcPaint.left,
                            lPaintStruct.rcPaint.top,
                            lPaintStruct.rcPaint.right - lPaintStruct.rcPaint.left,
                            lPaintStruct.rcPaint.bottom - lPaintStruct.rcPaint.top);

                        if(clipRect.Height >0 & clipRect.Width >0) 
                        {
                            // Double-buffer by painting everything to an image and then drawing the image.
                            int lWidth = (mMdiClient.ClientRectangle.Width > 0 ? mMdiClient.ClientRectangle.Width : 0);
                            int lHeight = (mMdiClient.ClientRectangle.Height > 0 ? mMdiClient.ClientRectangle.Height : 0);

                            using (Image lImage = new Bitmap(lWidth, lHeight)) 
                            {
                                using (Graphics lGraphics = Graphics.FromImage(lImage)) 
                                {
                                    // Draw base graphics and raise the base Paint event.
                                    IntPtr hdc = lGraphics.GetHdc();
                                    Message printClientMessage = Message.Create(Msg.HWnd, WM_PRINTCLIENT, hdc, IntPtr.Zero);
                                    DefWndProc(ref printClientMessage);
                                    lGraphics.ReleaseHdc(hdc);

                                    // Draw the image here.
                                    if (mImage != null)
                                        DrawImage(lGraphics, clipRect);

                                    // Call our OnPaint here to draw graphics over the
                                    // original and raise our Paint event.
                                    OnPaint(new PaintEventArgs(lGraphics, clipRect));
                                }

                                // Now draw all the graphics at once.
                                screenGraphics.DrawImage(lImage, mMdiClient.ClientRectangle);
                            }
                        
                        }
                    }

                    EndPaint(Msg.HWnd, ref lPaintStruct);
                    return;

                case WM_SIZE:
                    // Repaint on every resize.
                    mMdiClient.Invalidate();
                    break;

                case WM_NCCALCSIZE:
                    // If AutoScroll is set to false, hide the scrollbars when the control
                    // calculates its non-client area.
                    if (!mAutoScroll)
                        ShowScrollBar(Msg.HWnd, SB_BOTH, 0 /*false*/);
                    break;
            }

            base.WndProc(ref Msg);
        }

        /// <summary>
        /// Raises the <see cref="MdiClientController.Paint"/> event.
        /// </summary>
        /// <param name="EventArgs">A <see cref="System.Windows.Forms.PaintEventArgs"/> that
        /// contains the event data.</param>
        protected virtual void OnPaint(PaintEventArgs EventArgs)
        {
            // Raise the Paint event.
            if (Paint != null)
                Paint(this, EventArgs);
        }

        /// <summary>
        /// Raises the <see cref="MdiClientController.HandleAssigned"/> event.
        /// </summary>
        /// <param name="e">A <see cref="System.EventArgs"/> that contains the event
        /// data.</param>
        protected virtual void OnHandleAssigned(EventArgs e)
        {
            // Raise the HandleAssigned event.
            if (HandleAssigned != null)
                HandleAssigned(this, e);
        }
        #endregion

        #region Private Methods
        private void InitializeMdiClient()
        {
            // If the mdiClient has previously been set, unwire events connected to the old MDI.
            if (mMdiClient != null)
                mMdiClient.HandleDestroyed -= new EventHandler(MdiClientHandleDestroyed);

            if (mParentMDIForm == null)
                return;

            // Get the MdiClient from the parent form.
            for (int i = 0; i < mParentMDIForm.Controls.Count; i++)
            {
                // If the form is an MDI container, it will contain an MdiClient control
                // just as it would any other control.
                mMdiClient = mParentMDIForm.Controls[i] as MdiClient;
                if (mMdiClient != null)
                {
                    // Assign the MdiClient Handle to the NativeWindow.
                    ReleaseHandle();
                    AssignHandle(mMdiClient.Handle);

                    // Raise the HandleAssigned event.
                    OnHandleAssigned(EventArgs.Empty);

                    // Monitor the MdiClient for when its handle is destroyed.
                    mMdiClient.HandleDestroyed += new EventHandler(MdiClientHandleDestroyed);
                }
            }
        }

        private void DrawImage(Graphics GraphicsObj, Rectangle ClipRect)
        {
            // Paint the background image.
            if (stretchImage)
                GraphicsObj.DrawImage(mImage, mMdiClient.ClientRectangle);
            else
            {
                // Calculate the location of the image. (Note: this logic could be calculated during sizing
                // instead of during painting to improve performance.)
                Point lPoint = Point.Empty;
                switch (mImageAlignment)
                {
                    case ContentAlignment.TopLeft:
                        lPoint = new Point(0, 0);
                        break;
                    case ContentAlignment.TopCenter:
                        lPoint = new Point((mMdiClient.ClientRectangle.Width / 2) - (mImage.Width / 2), 0);
                        break;
                    case ContentAlignment.TopRight:
                        lPoint = new Point(mMdiClient.ClientRectangle.Width - mImage.Width, 0);
                        break;
                    case ContentAlignment.MiddleLeft:
                        lPoint = new Point(0,
                            (mMdiClient.ClientRectangle.Height / 2) - (mImage.Height / 2));
                        break;
                    case ContentAlignment.MiddleCenter:
                        lPoint = new Point((mMdiClient.ClientRectangle.Width / 2) - (mImage.Width / 2),
                            (mMdiClient.ClientRectangle.Height / 2) - (mImage.Height / 2));
                        break;
                    case ContentAlignment.MiddleRight:
                        lPoint = new Point(mMdiClient.ClientRectangle.Width - mImage.Width,
                            (mMdiClient.ClientRectangle.Height / 2) - (mImage.Height / 2));
                        break;
                    case ContentAlignment.BottomLeft:
                        lPoint = new Point(0, mMdiClient.ClientRectangle.Height - mImage.Height);
                        break;
                    case ContentAlignment.BottomCenter:
                        lPoint = new Point((mMdiClient.ClientRectangle.Width / 2) - (mImage.Width / 2),
                            mMdiClient.ClientRectangle.Height - mImage.Height);
                        break;
                    case ContentAlignment.BottomRight:
                        lPoint = new Point(mMdiClient.ClientRectangle.Width - mImage.Width,
                            mMdiClient.ClientRectangle.Height - mImage.Height);
                        break;
                }

                // Paint the image with the calculated coordinates and image size.
                GraphicsObj.DrawImage(mImage, new Rectangle(lPoint, mImage.Size));
            }
        }

        private void UpdateStyles()
        {
            // To show style changes, the non-client area must be repainted. Using the
            // control's Invalidate method does not affect the non-client area.
            // Instead use a Win32 call to signal the style has changed.

            SetWindowPos(mMdiClient.Handle, IntPtr.Zero, 0, 0, 0, 0,
                SWP_NOACTIVATE | SWP_NOMOVE | SWP_NOSIZE | SWP_NOZORDER |
                SWP_NOOWNERZORDER | SWP_FRAMECHANGED);
        }

        private void MdiClientHandleDestroyed(object Sender, EventArgs EventArgs)
        {
            // If the MdiClient handle has been released, drop the reference and
            // release the handle.
            if (mMdiClient != null)
            {
                mMdiClient.HandleDestroyed -= new EventHandler(MdiClientHandleDestroyed);
                mMdiClient = null;
            }

            ReleaseHandle();
        }

        private void ParentFormHandleCreated(object Sender, EventArgs EventArgs)
        {
            // The form has been created, unwire the event, and initialize the MdiClient.
            mParentMDIForm.HandleCreated -= new EventHandler(ParentFormHandleCreated);
            InitializeMdiClient();
            RefreshProperties();
        }

        private void RefreshProperties()
        {
            // Refresh all the properties
            BackColor = mBackColor;
            BorderStyle = mBorderStyle;
            AutoScroll = mAutoScroll;
            Image = mImage;
            ImageAlign = mImageAlignment;
            StretchImage = stretchImage;
        }
        #endregion

        #region Win32 API
        private const int WM_PAINT = 0x000F;
        private const int WM_ERASEBKGND = 0x0014;
        private const int WM_NCPAINT = 0x0085;
        private const int WM_THEMECHANGED = 0x031A;
        private const int WM_NCCALCSIZE = 0x0083;
        private const int WM_SIZE = 0x0005;
        private const int WM_PRINTCLIENT = 0x0318;

        private const uint SWP_NOSIZE = 0x0001;
        private const uint SWP_NOMOVE = 0x0002;
        private const uint SWP_NOZORDER = 0x0004;
        private const uint SWP_NOREDRAW = 0x0008;
        private const uint SWP_NOACTIVATE = 0x0010;
        private const uint SWP_FRAMECHANGED = 0x0020;
        private const uint SWP_SHOWWINDOW = 0x0040;
        private const uint SWP_HIDEWINDOW = 0x0080;
        private const uint SWP_NOCOPYBITS = 0x0100;
        private const uint SWP_NOOWNERZORDER = 0x0200;
        private const uint SWP_NOSENDCHANGING = 0x0400;

        private const int WS_BORDER = 0x00800000;
        private const int WS_EX_CLIENTEDGE = 0x00000200;
        private const int WS_DISABLED = 0x08000000;

        private const int GWL_STYLE = -16;
        private const int GWL_EXSTYLE = -20;

        private const int SB_HORZ = 0;
        private const int SB_VERT = 1;
        private const int SB_CTL = 2;
        private const int SB_BOTH = 3;

        [StructLayout(LayoutKind.Sequential)]
        private struct RECT
        {
            public int left;
            public int top;
            public int right;
            public int bottom;

            public RECT(Rectangle rect)
            {
                this.left = rect.Left;
                this.top = rect.Top;
                this.right = rect.Right;
                this.bottom = rect.Bottom;
            }

            public RECT(int left, int top, int right, int bottom)
            {
                this.left = left;
                this.top = top;
                this.right = right;
                this.bottom = bottom;
            }
        }

        [StructLayout(LayoutKind.Sequential, Pack = 4)]
        private struct PAINTSTRUCT
        {
            public IntPtr hdc;
            public int fErase;
            public RECT rcPaint;
            public int fRestore;
            public int fIncUpdate;
            [MarshalAs(UnmanagedType.ByValArray, SizeConst = 32)]
            public byte[] rgbReserved;
        }

        [StructLayout(LayoutKind.Sequential)]
        private struct NCCALCSIZE_PARAMS
        {
            public RECT rgrc0, rgrc1, rgrc2;
            public IntPtr lppos;
        }

        [DllImport("user32.dll")]
        private static extern int ShowScrollBar(IntPtr hWnd, int wBar, int bShow);

        [DllImport("user32.dll")]
        private static extern IntPtr BeginPaint(IntPtr hWnd, ref PAINTSTRUCT paintStruct);

        [DllImport("user32.dll")]
        private static extern bool EndPaint(IntPtr hWnd, ref PAINTSTRUCT paintStruct);

        [DllImport("user32.dll", CharSet = CharSet.Auto)]
        private static extern int GetWindowLong(IntPtr hWnd, int Index);

        [DllImport("user32.dll", CharSet = CharSet.Auto)]
        private static extern int SetWindowLong(IntPtr hWnd, int Index, int Value);

        [DllImport("user32.dll", ExactSpelling = true)]
        private static extern int SetWindowPos(IntPtr hWnd, IntPtr hWndInsertAfter, int X, int Y, int cx, int cy, uint uFlags);
        #endregion
    }
}
