﻿using Microsoft.Win32.SafeHandles;
using System;
using System.ComponentModel;
using System.Drawing;
using System.Runtime.InteropServices;
using System.Windows.Forms;

namespace Eka.Inheritance.MDIControls.FormSearchBox
{
    /// <summary>
    /// A dropdown window for combos.
    /// </summary>
    [ToolboxItem(false)]
    public class CustomDropDown : ToolStripDropDown, IMessageFilter
    {
        #region Member variables
        private const int WM_LBUTTONDOWN = 0x0201;
        private const int WM_RBUTTONDOWN = 0x0204;
        private const int WM_MBUTTONDOWN = 0x0207;
        private const int WM_NCLBUTTONDOWN = 0x00A1;
        private const int WM_NCRBUTTONDOWN = 0x00A4;
        private const int WM_NCMBUTTONDOWN = 0x00A7;

        /// <summary>
        /// 
        /// </summary>
        /// <param name="hWndFrom"></param>
        /// <param name="hWndTo"></param>
        /// <param name="pt"></param>
        /// <param name="cPoints"></param>
        /// <returns></returns>
        [DllImport("user32.dll", ExactSpelling = true, CharSet = CharSet.Auto)]
        public static extern int MapWindowPoints(IntPtr hWndFrom, IntPtr hWndTo, [In, Out] ref Point pt, int cPoints);

        /// <summary>
        /// member variable to host the tooltrip custom control
        /// </summary>
        private ToolStripControlHost mHost;

        /// <summary>
        /// member variable for control, which we display this dropdown
        /// </summary>
        private Control mOpenerControl;

        /// <summary>
        /// Flag: Has Dispose already been called?
        /// </summary>
        bool mDisposed = false;

        /// <summary>
        /// Instantiate a SafeHandle instance.
        /// </summary>
        SafeHandle mSafeHandle = new SafeFileHandle(IntPtr.Zero, true);
        #endregion

        #region Constructor
        /// <summary>
        /// Constructor
        /// </summary>
        public CustomDropDown(Control ContentControl)
        {
            if (ContentControl == null)
            {
                throw new ArgumentNullException("ContentControl");
            }

            Content = ContentControl;
            Content.Location = Point.Empty;
            mHost = new ToolStripControlHost(Content);
            // AutoClose must be set to false, because otherwise the ToolStripManager would steal keyboard events
            AutoClose = false;

            // make autosize to false so that we will do the sizing in the code
            AutoSize = false;
            DoubleBuffered = true;
            ResizeRedraw = false;
            Padding = Margin = mHost.Padding = mHost.Margin = Padding.Empty;

            // we adjust the size according to the contents
            MinimumSize = Content.MinimumSize;
            ContentControl.MinimumSize = Content.Size;
            MaximumSize = Content.MaximumSize;
            ContentControl.MaximumSize = Content.Size;
            Size = Content.Size;
            TabStop = Content.TabStop = true;

            // set up the content
            Items.Add(mHost);

            // we must listen to mouse events for "emulating" AutoClose
            Application.AddMessageFilter(this);
        }
        #endregion

        #region Properties
        /// <summary>
        /// Gets the content of the pop-up.
        /// </summary>
        public Control Content
        {
            get;
            private set;
        }
        #endregion

        #region Public Methods
        /// <summary>
        /// Display the dropdown and adjust its size and location
        /// </summary>
        public void Show(Control OpenerControl, Size PreferredSize = new Size())
        {
            if (OpenerControl == null)
            {
                throw new ArgumentNullException("OpenerControl");
            }

            mOpenerControl = OpenerControl;

            int lWidth = PreferredSize.Width == 0 ? ClientRectangle.Width : PreferredSize.Width;
            int lHeight = PreferredSize.Height == 0 ? Content.Height : PreferredSize.Height;
            lHeight += Padding.Size.Height + Content.Margin.Size.Height;

            Rectangle screen = Screen.FromControl(mOpenerControl).WorkingArea;

            // Try first to place it below the opener control
            Rectangle lRectangleLocation = mOpenerControl.RectangleToScreen(
                    new Rectangle(mOpenerControl.ClientRectangle.Left, mOpenerControl.ClientRectangle.Bottom,
                    mOpenerControl.ClientRectangle.Left + lWidth, mOpenerControl.ClientRectangle.Bottom + lHeight));

            Point lLocation = new Point(mOpenerControl.ClientRectangle.Left, mOpenerControl.ClientRectangle.Bottom);
            if (!screen.Contains(lRectangleLocation))
            {
                // let's try above the opener control
                lRectangleLocation = mOpenerControl.RectangleToScreen(
                        new Rectangle(mOpenerControl.ClientRectangle.Left, mOpenerControl.ClientRectangle.Top - lHeight,
                        mOpenerControl.ClientRectangle.Left + lWidth, mOpenerControl.ClientRectangle.Top));

                if (screen.Contains(lRectangleLocation))
                {
                    lLocation = new Point(mOpenerControl.ClientRectangle.Left, mOpenerControl.ClientRectangle.Top - lHeight);
                }
            }

            Width = lWidth;
            Height = lHeight;
            Show(mOpenerControl, lLocation, ToolStripDropDownDirection.BelowRight);
        }

        #region IMessageFilter implementation
        /// <summary>
        ///We monitor all messages in order to detect when the users clicks outside the dropdown and the combo
        /// If this happens, we close the dropdown (as AutoClose is false)
        /// </summary>
        /// <param name="Msg">Message</param>
        /// <returns></returns>
        public bool PreFilterMessage(ref Message Msg)
        {
            if (Visible)
            {
                switch (Msg.Msg)
                {
                    case WM_LBUTTONDOWN:
                    case WM_RBUTTONDOWN:
                    case WM_MBUTTONDOWN:
                    case WM_NCLBUTTONDOWN:
                    case WM_NCRBUTTONDOWN:
                    case WM_NCMBUTTONDOWN:
                        int lLParam = unchecked((int)(long)Msg.LParam);
                        short lXPosition = (short)(lLParam & 0xFFFF);
                        short lYPosition = (short)((lLParam >> 16) & 0xffff);
                        Point lPoint = new Point(lXPosition, lYPosition);
                        IntPtr lSrcWnd =
                            // client area: x, y are relative to the client area of the windows
                            (Msg.Msg == WM_LBUTTONDOWN) || (Msg.Msg == WM_RBUTTONDOWN) || (Msg.Msg == WM_MBUTTONDOWN) ?
                            Msg.HWnd :
                            // non-client area: x, y are relative to the desktop
                            IntPtr.Zero;

                        MapWindowPoints(lSrcWnd, Handle, ref lPoint, 1);
                        if (!ClientRectangle.Contains(lPoint))
                        {
                            // the user has clicked outside the dropdown
                            lPoint = new Point(lXPosition, lYPosition);
                            MapWindowPoints(lSrcWnd, mOpenerControl.Handle, ref lPoint, 1);
                            if (!mOpenerControl.ClientRectangle.Contains(lPoint))
                            {
                                // the user has clicked outside the opener control
                                Close();
                            }
                        }
                        break;
                }
            }
            return false;
        }
        #endregion
        #endregion

        #region Protected methods
        /// <summary>
        /// On resizes, resize the contents
        /// </summary>
        protected override void OnSizeChanged(EventArgs e)
        {
            if (Content != null)
            {
                Content.MinimumSize = Size;
                Content.MaximumSize = Size;
                Content.Size = Size;
                Content.Location = Point.Empty;
            }
            base.OnSizeChanged(e);
        }

        /// <summary>
        /// Clean up any resources being used.
        /// </summary>
        /// <param name="Disposing"></param>
        protected override void Dispose(bool Disposing)
        {
            if (mDisposed)
                return;

            if (Disposing)
            {
                mSafeHandle.Dispose();
                mSafeHandle = null;

                // Free any other managed objects here.
                if (Content != null)
                {
                    Content.Dispose();
                    Content = null;
                }

                if (mHost != null)
                {
                    mHost.Dispose();
                    mHost = null;
                }

                //Do not dispose mOpenerControl, simply make it null
                mOpenerControl = null;

                Application.RemoveMessageFilter(this);
            }

            // Free any unmanaged objects here.below

            //set the value to indicate the object has been disposed
            mDisposed = true;

            // Call base class implementation.
            base.Dispose(Disposing);
        }
        #endregion
    }
}
