﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Data;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;
using System.Runtime.InteropServices;
using System.Collections;
using Microsoft.Win32.SafeHandles;

namespace Eka.Inheritance.MDIControls.FormSearchBox
{
    /// <summary>
    /// Custom control to search items using search type(wild card/regular exp) from the control.
    /// </summary>
    public partial class FormSearchBox : UserControl
    {
        #region Member variables and constant
        private const int TV_FIRST = 0x1100;
        private const int TVM_SETBKCOLOR = TV_FIRST + 29;
        private const int TVM_SETEXTENDEDSTYLE = TV_FIRST + 44;
        private const int TVS_EX_DOUBLEBUFFER = 0x0004;

        /// <summary>
        /// Win API to set DoubleBuffered = true for the treeview to avoid flickring
        /// </summary>
        /// <param name="hWnd"></param>
        /// <param name="msg"></param>
        /// <param name="wParam"></param>
        /// <param name="lParam"></param>
        /// <returns></returns>
        [DllImport("user32.dll")]
        public static extern IntPtr SendMessage(IntPtr hWnd, int msg, IntPtr wParam, IntPtr lParam);

        /// <summary>
        /// Memeber variable to holds the Controls Border color (#666666)
        /// </summary>
        private Color mBorderColor = Color.LightGray; //Color.FromArgb(((int)(((byte)(102)))), ((int)(((byte)(102)))), ((int)(((byte)(102)))));
               
        /// <summary>
        /// An event that clients can use to be notified whenever the elements of the list change.
        /// </summary>
        public event AutoCompleComboBox.SearchTextHandler SearchItemSelected;

        /// <summary>
        /// Flag: Has Dispose already been called?
        /// </summary>
        bool mDisposed = false;

        /// <summary>
        /// Instantiate a SafeHandle instance.
        /// </summary>
        SafeHandle mSafeHandle = new SafeFileHandle(IntPtr.Zero, true);
        #endregion

        #region Constructor
        /// <summary>
        /// Constructor
        /// </summary>
        public FormSearchBox()
        {
            InitializeComponent();

            // Enable default double buffering processing
            SetStyle(ControlStyles.OptimizedDoubleBuffer | ControlStyles.AllPaintingInWmPaint, true);

            // Disable default CommCtrl painting on non-Vista systems
            if (!IsWinVista)
                SetStyle(ControlStyles.UserPaint, true);

            //Initialize the basic control theme
            InitializeTheme();

            //Reroute the Selected Item selected change event
            CtrAutoCompleteComboBox.SearchItemSelected += OnSearchItemSelected;
        }
        #endregion

        #region Properties
        /// <summary>
        /// Do not allow to set the border style for this Control
        /// </summary>
        [DefaultValue(BorderStyle.None)]
        public new BorderStyle BorderStyle
        {
            get
            {
                return base.BorderStyle;
            }
            set { }
        }

        /// <summary>
        /// Get collection of search item binded in the control
        /// </summary>
        public ComboBox.ObjectCollection Items
        {
            get
            {
                return CtrAutoCompleteComboBox.Items;
            }
        }

        /// <summary>
        /// Get/Set control border color
        /// </summary>
        public Color BorderColor
        {
            get
            {
                return CtrAutoCompleteComboBox.BorderColor; 
            }

            set
            {
                CtrAutoCompleteComboBox.BorderColor = value; 
            }
        }

        /// <summary>
        /// Get/Set control's floating text
        /// </summary>
        public string FloatingText
        {
            get
            {
                return CtrAutoCompleteComboBox.FloatingText;
            }

            set
            {
                CtrAutoCompleteComboBox.FloatingText = value; 
            }
        }

        /// <summary>
        /// Get/Set value for drawing border. 
        /// True: draw border, which is default value.
        /// False: without border
        /// </summary>
        [DefaultValue(true)]
        public bool DrawBorder
        {
            get
            {
                return CtrAutoCompleteComboBox.DrawBorder; 
            }
            set
            {
                CtrAutoCompleteComboBox.DrawBorder = value;
            }
        }
        
        /// <summary>
        /// Get/Set the forecolor of editable control text
        /// </summary>
        public Color DropDownForeColor
        {
            get
            {
                return CtrAutoCompleteComboBox.ForeColor; 
            }
            set
            {
                CtrAutoCompleteComboBox.ForeColor = value;
            }
        }
        #endregion

        #region Public Method
        #endregion

        #region Procted Method
        /// <summary>
        /// Override the OnHandleCreated Method
        /// </summary>
        /// <param name="e"></param>
        protected override void OnHandleCreated(EventArgs e)
        {
            base.OnHandleCreated(e);

            UpdateExtendedStyles();
            if (!IsWinXP)
                SendMessage(Handle, TVM_SETBKCOLOR, IntPtr.Zero, (IntPtr)ColorTranslator.ToWin32(BackColor));
        }

        /// <summary>
        /// override the OnResize method
        /// </summary>
        /// <param name="e"></param>
        protected override void OnResize(EventArgs e)
        {
            base.OnResize(e);
            ResizeControls();
            Invalidate();
        }
        #endregion

        #region Private Method
        /// <summary>
        /// Method to initialze the basic theme of the control
        /// </summary>
        private void InitializeTheme()
        {
            //Set the padding
            Padding = new Padding(0, 0, 0, 0);

            //Set background color as white by default
            BackColor = Color.White;

            //Initialize the control default size: default width-200 and height= height of inner combobox
            Size = new System.Drawing.Size(200, CalculateControlHeight());

            //always set the border style as None, so that we wont be able to draw the border
            BorderStyle = System.Windows.Forms.BorderStyle.None;

            //Put the search picture box always in top, so that it will hide the default combobox dropwown image        
            ctrPictureBox.BringToFront();

            //Resizie the internal constrols so that it will exactly fit in tis user control
            ResizeControls();

            //set the properties value for internal custom combobox
            CtrAutoCompleteComboBox.BorderColor = Color.LightGray;
            CtrAutoCompleteComboBox.MatchingType = SearchMatchingType.UseWildcard;
            CtrAutoCompleteComboBox.Dock = DockStyle.Fill;
            CtrAutoCompleteComboBox.BackColor = Color.White;
            CtrAutoCompleteComboBox.ForeColor = Color.Black; 

            //Set the PIcturebox theme
            ctrPictureBox.Dock = DockStyle.Right;
            ctrPictureBox.BackColor = Color.Transparent;
            ctrPictureBox.BorderStyle = BorderStyle.None;
            ctrPictureBox.SizeMode = PictureBoxSizeMode.StretchImage;           
        }

        /// <summary>
        /// Raise search item chagne event
        /// </summary>
        /// <param name="eventArgs"></param>
        private void OnSearchItemSelected(SearchTextArgs eventArgs)
        {
            if (SearchItemSelected == null)
                return;

            SearchItemSelected(eventArgs);

        }

        /// <summary>
        /// Calculate control height
        /// </summary>
        /// <returns></returns>
        private int CalculateControlHeight()
        {
            return CtrAutoCompleteComboBox.Height;
        }

        /// <summary>
        /// Method to resize the controls when textbox size changed
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void OnEkaTextBoxResize(object sender, EventArgs e)
        {
            ResizeControls();
        }

        /// <summary>
        /// Method to calculate size and position to place the child controls in proper place'
        /// and set the size
        /// </summary>
        private void ResizeControls()
        {
            Height = CalculateControlHeight();
            ctrPictureBox.Size = new Size(CtrAutoCompleteComboBox.Height, CtrAutoCompleteComboBox.Height);
        }

        private void UpdateExtendedStyles()
        {
            int lStyle = 0;

            if (DoubleBuffered)
                lStyle |= TVS_EX_DOUBLEBUFFER;

            if (lStyle != 0)
                SendMessage(Handle, TVM_SETEXTENDEDSTYLE, (IntPtr)TVS_EX_DOUBLEBUFFER, (IntPtr)lStyle);
        }

        private static bool IsWinVista
        {
            get
            {
                OperatingSystem OS = Environment.OSVersion;
                return (OS.Platform == PlatformID.Win32NT) && (OS.Version.Major >= 6);
            }
        }

        private static bool IsWinXP
        {
            get
            {
                OperatingSystem OS = Environment.OSVersion;
                return (OS.Platform == PlatformID.Win32NT) &&
                    ((OS.Version.Major > 5) || ((OS.Version.Major == 5) && (OS.Version.Minor == 1)));
            }
        }        
        #endregion

        #region Dispose object
        /// <summary>
        ///  Protected implementation of Dispose pattern.
        /// </summary>
        /// <param name="Disposing"></param>
        protected override void Dispose(bool Disposing)
        {
            if (mDisposed)
                return;

            if (Disposing)
            {
                mSafeHandle.Dispose();
                mSafeHandle = null;

                // Free any other managed objects here.
                if (CtrAutoCompleteComboBox != null)
                {
                    CtrAutoCompleteComboBox.SearchItemSelected -= OnSearchItemSelected;
                    CtrAutoCompleteComboBox.Dispose();
                    CtrAutoCompleteComboBox = null;
                }

                if (ctrPictureBox != null)
                {
                    ctrPictureBox.Dispose();
                    ctrPictureBox = null;
                }

                SearchItemSelected = null;
            }

            // Free any unmanaged objects here.

            mDisposed = true;
            // Call base class implementation.
            base.Dispose(Disposing);
        }
        #endregion
    }    
}
