﻿using Microsoft.Win32.SafeHandles;
using System;
using System.Collections.Generic;
using System.Runtime.InteropServices;
using System.Text;
using System.Text.RegularExpressions;

namespace Eka.Inheritance.MDIControls.FormSearchBox
{
    /// <summary>
    /// Allows to decompose a strings against a pattern
    /// </summary>
    public class SearchStringMatcher : IDisposable
    {
        #region Member variables
        /// <summary>
        /// Member variable to holds the search patterns object
        /// </summary>
        private object mSearchPattern;

        /// <summary>
        /// Flag: Has Dispose already been called?
        /// </summary>
        bool mDisposed = false;

        /// <summary>
        /// Instantiate a SafeHandle instance.
        /// </summary>
        SafeHandle mSafeHandle = new SafeFileHandle(IntPtr.Zero, true);

        /// <summary>
        /// Delegate to the search matching type
        /// </summary>
        /// <param name="SourceString"></param>
        /// <returns></returns>
        public delegate SearchMatchdInfo MatchDelegate(string SourceString);
        #endregion

        /// <summary>
        /// Constructor
        /// </summary>
        public SearchStringMatcher(SearchMatchingType MatchingType, string SearchPatternString)
        {
            switch (MatchingType)
            {
                case SearchMatchingType.UseRegularExpression:
                    //Prepare regulara expression search
                    mSearchPattern = PrepareRegularExpression(SearchPatternString);
                    Match = BuildRegularExpression;
                    break;

                case SearchMatchingType.UseWildcard:
                    mSearchPattern = PrepareWithWildcards(SearchPatternString);
                    Match = BuildWithWildcards;
                    break;

                case SearchMatchingType.NoWildcard:
                    mSearchPattern = PrepareNoWildcard(SearchPatternString);
                    Match = BuildNoWildcard;
                    break;

                default:
                    throw new ArgumentException("Unknown StringMatchingMethod");
            }
        }

        /// <summary>
        /// Match against a string
        /// </summary>
        public MatchDelegate Match
        {
            get;
            private set;
        }

        #region Private Methods
        #region No wildcards
        /// <summary>
        /// Prepare the pattern for a "non wildcard" match
        /// </summary>
        private object PrepareNoWildcard(string SearchPattern)
        {
            return SearchPattern.ToLower();
        }

        /// <summary>
        /// Compare the source against the pattern and if successfull returns a StringMatch
        /// 
        /// There is a match if source contains all the characters of pattern in the right order
        /// but not consecutively
        /// </summary>
        private SearchMatchdInfo BuildNoWildcard(string SourceString)
        {
            string lPattern = (string)mSearchPattern;
            string lSource = SourceString.ToLower();
            List<string> lSearchSegments = new List<string>();
            int lIndexi = 0, lIndexj;
            int NP = lPattern.Length, NS = SourceString.Length;
            bool lStartsOnMatch = false, lIsMatch = false;

            for (lIndexj = 0; lIndexj < NP; lIndexj++)
            {
                int s = lIndexi;
                // skip characters until we have a match
                for (; lIndexi < NS; lIndexi++)
                {
                    if (lSource[lIndexi] == lPattern[lIndexj])
                    {
                        lIsMatch = (lIndexj == NP - 1);
                        if (s != lIndexi)
                        {
                            lSearchSegments.Add(SourceString.Substring(s, lIndexi - s));
                            lSearchSegments.Add(SourceString.Substring(lIndexi, 1));
                        }
                        else
                        {
                            if (lSearchSegments.Count == 0)
                            {
                                lSearchSegments.Add(SourceString.Substring(lIndexi, 1));
                                lStartsOnMatch = true;
                            }
                            else
                            {
                                lSearchSegments[lSearchSegments.Count - 1] += SourceString[lIndexi];
                            }
                        }
                        lIndexi++;
                        break; // i loop
                    }
                }
                if (lIndexi >= NS)
                {
                    break; // j loop
                }
            }

            if (!lIsMatch)
            {
                return null;
            }
            if (lIndexi < NS)
            {
                lSearchSegments.Add(SourceString.Substring(lIndexi, NS - lIndexi));
            }
            return new SearchMatchdInfo
            {
                Text = SourceString,
                Segments = lSearchSegments,
                StartsOnMatch = lStartsOnMatch,
            };
        }
        #endregion

        #region Wildcards
        /// <summary>
        /// Prepare the pattern for a "wildcard" match
        /// </summary>
        private object PrepareWithWildcards(string SearchPatternString)
        {
            return SearchPatternString;
        }       

        /// <summary>
        ///  Compare the source against the pattern and if successfull returns a StringMatch.
        ///  There match is done against the source which can contain the wildcard '*'
        /// </summary>
        /// <param name="SourceString">source string where to be search</param>
        /// <returns></returns>
        private SearchMatchdInfo BuildWithWildcards(string SourceString)
        {
            SearchMatchdInfo lSearchMatch = new SearchMatchdInfo
            {
                Text = SourceString,
                Segments = new List<string>(),
                StartsOnMatch = false,
            };

            int lFirstIndex = SourceString.ToUpper().IndexOf(mSearchPattern.ToString().ToUpper());

            if (lFirstIndex < 0)
                return null;

            int lSearchTextLen = mSearchPattern.ToString().Length;

            if (SourceString.ToUpper().StartsWith(mSearchPattern.ToString().ToUpper()))
            {
                lSearchMatch.Segments.Add(SourceString.Substring(0, lSearchTextLen));
                lSearchMatch.Segments.Add(SourceString.Remove(0, lSearchTextLen));
                lSearchMatch.StartsOnMatch = true;
            }
            else
            {
                lSearchMatch.Segments.Add(SourceString.Substring(0, lFirstIndex));
                lSearchMatch.Segments.Add(SourceString.Substring(lFirstIndex, mSearchPattern.ToString().Length));
                string lEndSegment = SourceString.Remove(0, lFirstIndex + lSearchTextLen);
                if (lEndSegment.Length > 0)
                    lSearchMatch.Segments.Add(SourceString.Remove(0, lFirstIndex + lSearchTextLen));
            }

            if (lSearchMatch.Segments.Count == 0)
                return null;

            return lSearchMatch;
        }
        #endregion

        #region Regular Expression
        /// <summary>
        /// Prepare the pattern for a Regular Expression match
        /// </summary>
        /// <param name="SearchPatternString">The regular expression pattern to match</param>
        /// <returns>the string object if it is valid</returns>
        private static object PrepareRegularExpression(string SearchPatternString)
        {
            try
            {
                return new Regex(SearchPatternString, RegexOptions.IgnoreCase);
            }
            catch
            {
                // the user has entered an invalid regex, just ignore it
            }
            return null;
        }

        /// <summary>
        /// Compare the source against the pattern and if successfull returns a StringMatch.
        /// 
        /// There is a match if source contains all the characters of pattern in the right order but not consecutively.
        /// </summary>
        /// <param name="SourceString">Source String where to be search</param>
        /// <returns>SearchMatch</returns>
        private SearchMatchdInfo BuildRegularExpression(string SourceString)
        {
            if (mSearchPattern == null) return null;

            Regex lRegex = (Regex)mSearchPattern;
            List<string> lSegmentList = new List<string>();
            bool lStartsOnMatch = false;
            bool lIsMatch = false;
            int lIndex = 0;

            // we need only one match
            Match lMatch = lRegex.Match(SourceString, lIndex);

            if (lMatch.Success)
            {
                lIsMatch = true;
                if (lMatch.Index == lIndex)
                {
                    if (lMatch.Index == 0)
                    {
                        lStartsOnMatch = true;
                    }
                    lSegmentList.Add(SourceString.Substring(lMatch.Index, lMatch.Length));
                }
                else
                {
                    lSegmentList.Add(SourceString.Substring(lIndex, lMatch.Index - lIndex));
                    lSegmentList.Add(SourceString.Substring(lMatch.Index, lMatch.Length));
                }
                lIndex = lMatch.Index + lMatch.Length;
            }

            if (!lIsMatch) return null;

            if (lIndex < SourceString.Length)
            {
                lSegmentList.Add(SourceString.Substring(lIndex, SourceString.Length - lIndex));
            }

            return new SearchMatchdInfo
            {
                Text = SourceString,
                Segments = lSegmentList,
                StartsOnMatch = lStartsOnMatch,
            };
        }
        #endregion
        #endregion

        #region Dispose object
        /// <summary>
        ///  Public implementation of Dispose pattern callable by consumers.
        /// </summary>
        public void Dispose()
        {
            Dispose(true);
            GC.SuppressFinalize(this);
        }

        /// <summary>
        /// Protected implementation of Dispose pattern.
        /// </summary>
        /// <param name="Disposing"></param>
        protected virtual void Dispose(bool Disposing)
        {
            if (mDisposed)
                return;

            if (Disposing)
            {
                mSafeHandle.Dispose();
                mSafeHandle = null;

                // Free any other managed objects here.
                mSearchPattern = null;
                Match = null;
            }

            // Free any unmanaged objects here. 

            //Finally inform GC object has been disposed as mDisposed = true
            mDisposed = true;
        }
        #endregion
    }   
}
