﻿using System;
using System.Drawing;
using System.Windows.Forms;

namespace Eka.Inheritance.MDIControls
{
    /// <summary>
    /// Abstract Popup Control class
    /// This allows the construction of quick pop-up controls.
    /// </summary>
    public partial class ctrPopUp<T> : UserControl where T : Control, new()
    {
        const int mReopenDelay = 200;

        /// <summary>
        /// The actual drop down that shows the CheckedListBox
        /// </summary>
        protected ToolStripDropDown DropDown = new ToolStripDropDown();

        /// <summary>
        /// The Toolstrip Host
        /// (This can't be set until the control is ready)
        /// </summary>
        protected ToolStripControlHost lHost;

        /// <summary>
        /// The content itself
        /// </summary>
        protected T Content = new T();

        /// <summary>
        /// Used to prevent re-open of dropdown if the old one has just closed.
        /// For example, if this is a drop down list, then clicking on the
        /// list control (button with a drop triangle) will open the list.
        /// Clicking on it again should close the drop list, not re-open
        /// it again.
        /// </summary>
        private DateTime lIgnoreOpenUntil = DateTime.MinValue;

        public ctrPopUp()
        {
            InitializeComponent();

            // Prepare the dropdown for action
            lHost = new ToolStripControlHost(Content);

            DropDown.Items.Add(lHost);

            // Add event to prevent re-open when expecting close.
            DropDown.VisibleChanged += new EventHandler(dropDown_VisibleChanged);
        }

        #region Protected Methods
        /// <summary>
        /// Open the Pop up
        /// </summary>
        protected void Open()
        {
            // Prevent re-open of dropdown if the old one has just closed.
            // For example, if this is a drop down list, then clicking on the
            // list control (button with a drop triangle) will open the list.
            // Clicking on it again should close the drop list, not re-open
            // it again.
            if (DateTime.Now > lIgnoreOpenUntil)
            {
                lHost.Width = Content.Width;
                lHost.Margin = new Padding(0);
                lHost.Padding = new Padding(0);
                lHost.Height = Content.Height;
                DropDown.Show(this, 0, this.Height);
            }
        }
        #endregion

        #region Events
        /// <summary>
        /// Dropdown Visible changed.
        /// If it is now hidden, prevent immediate re-open
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void dropDown_VisibleChanged(object sender, EventArgs e)
        {
            Control c = sender as Control;
            if (c != null)
            {
                if (!c.Visible)
                {
                    lIgnoreOpenUntil = DateTime.Now.AddMilliseconds(mReopenDelay);
                }
            }
        }
        #endregion
    }
}